@@ -50,7 +50,7 @@ static const struct property_entry chuwi_hi8_air_props[] = {
 };
 
 static const struct ts_dmi_data chuwi_hi8_air_data = {
-	.acpi_name	= "MSSL1680",
+	.acpi_name	= "MSSL1680:00",
 	.properties	= chuwi_hi8_air_props,
 };
 
@@ -336,22 +336,6 @@ static const struct ts_dmi_data dexp_ursus_7w_data = {
 	.properties	= dexp_ursus_7w_props,
 };
 
-static const struct property_entry dexp_ursus_kx210i_props[] = {
-	PROPERTY_ENTRY_U32("touchscreen-min-x", 5),
-	PROPERTY_ENTRY_U32("touchscreen-min-y",  2),
-	PROPERTY_ENTRY_U32("touchscreen-size-x", 1720),
-	PROPERTY_ENTRY_U32("touchscreen-size-y", 1137),
-	PROPERTY_ENTRY_STRING("firmware-name", "/*(DEBLOBBED)*/"),
-	PROPERTY_ENTRY_U32("silead,max-fingers", 10),
-	PROPERTY_ENTRY_BOOL("silead,home-button"),
-	{ }
-};
-
-static const struct ts_dmi_data dexp_ursus_kx210i_data = {
-	.acpi_name	= "MSSL1680:00",
-	.properties	= dexp_ursus_kx210i_props,
-};
-
 static const struct property_entry digma_citi_e200_props[] = {
 	PROPERTY_ENTRY_U32("touchscreen-size-x", 1980),
 	PROPERTY_ENTRY_U32("touchscreen-size-y", 1500),
@@ -394,11 +378,6 @@ static const struct ts_dmi_data gdix1001_01_upside_down_data = {
 	.properties	= gdix1001_upside_down_props,
 };
 
-static const struct ts_dmi_data gdix1002_00_upside_down_data = {
-	.acpi_name	= "GDIX1002:00",
-	.properties	= gdix1001_upside_down_props,
-};
-
 static const struct property_entry gp_electronic_t701_props[] = {
 	PROPERTY_ENTRY_U32("touchscreen-size-x", 960),
 	PROPERTY_ENTRY_U32("touchscreen-size-y", 640),
@@ -740,21 +719,6 @@ static const struct ts_dmi_data pipo_w11_data = {
 	.properties	= pipo_w11_props,
 };
 
-static const struct property_entry positivo_c4128b_props[] = {
-	PROPERTY_ENTRY_U32("touchscreen-min-x", 4),
-	PROPERTY_ENTRY_U32("touchscreen-min-y", 13),
-	PROPERTY_ENTRY_U32("touchscreen-size-x", 1915),
-	PROPERTY_ENTRY_U32("touchscreen-size-y", 1269),
-	PROPERTY_ENTRY_STRING("firmware-name", "/*(DEBLOBBED)*/"),
-	PROPERTY_ENTRY_U32("silead,max-fingers", 10),
-	{ }
-};
-
-static const struct ts_dmi_data positivo_c4128b_data = {
-	.acpi_name	= "MSSL1680:00",
-	.properties	= positivo_c4128b_props,
-};
-
 static const struct property_entry pov_mobii_wintab_p800w_v20_props[] = {
 	PROPERTY_ENTRY_U32("touchscreen-min-x", 32),
 	PROPERTY_ENTRY_U32("touchscreen-min-y", 16),
@@ -854,21 +818,6 @@ static const struct ts_dmi_data rwc_nanote_p8_data = {
 	.properties = rwc_nanote_p8_props,
 };
 
-static const struct property_entry rwc_nanote_next_props[] = {
-	PROPERTY_ENTRY_U32("touchscreen-min-x", 5),
-	PROPERTY_ENTRY_U32("touchscreen-min-y", 5),
-	PROPERTY_ENTRY_U32("touchscreen-size-x", 1785),
-	PROPERTY_ENTRY_U32("touchscreen-size-y", 1145),
-	PROPERTY_ENTRY_BOOL("touchscreen-inverted-y"),
-	PROPERTY_ENTRY_STRING("firmware-name", "/*(DEBLOBBED)*/"),
-	{ }
-};
-
-static const struct ts_dmi_data rwc_nanote_next_data = {
-	.acpi_name = "MSSL1680:00",
-	.properties = rwc_nanote_next_props,
-};
-
 static const struct property_entry schneider_sct101ctm_props[] = {
 	PROPERTY_ENTRY_U32("touchscreen-size-x", 1715),
 	PROPERTY_ENTRY_U32("touchscreen-size-y", 1140),
@@ -886,22 +835,6 @@ static const struct ts_dmi_data schneider_sct101ctm_data = {
 	.properties	= schneider_sct101ctm_props,
 };
 
-static const struct property_entry globalspace_solt_ivw116_props[] = {
-	PROPERTY_ENTRY_U32("touchscreen-min-x", 7),
-	PROPERTY_ENTRY_U32("touchscreen-min-y", 22),
-	PROPERTY_ENTRY_U32("touchscreen-size-x", 1723),
-	PROPERTY_ENTRY_U32("touchscreen-size-y", 1077),
-	PROPERTY_ENTRY_STRING("firmware-name", "/*(DEBLOBBED)*/"),
-	PROPERTY_ENTRY_U32("silead,max-fingers", 10),
-	PROPERTY_ENTRY_BOOL("silead,home-button"),
-	{ }
-};
-
-static const struct ts_dmi_data globalspace_solt_ivw116_data = {
-	.acpi_name	= "MSSL1680:00",
-	.properties	= globalspace_solt_ivw116_props,
-};
-
 static const struct property_entry techbite_arc_11_6_props[] = {
 	PROPERTY_ENTRY_U32("touchscreen-min-x", 5),
 	PROPERTY_ENTRY_U32("touchscreen-min-y", 7),
@@ -944,32 +877,6 @@ static const struct ts_dmi_data teclast_tbook11_data = {
 	.properties	= teclast_tbook11_props,
 };
 
-static const struct property_entry teclast_x16_plus_props[] = {
-	PROPERTY_ENTRY_U32("touchscreen-min-x", 8),
-	PROPERTY_ENTRY_U32("touchscreen-min-y", 14),
-	PROPERTY_ENTRY_U32("touchscreen-size-x", 1916),
-	PROPERTY_ENTRY_U32("touchscreen-size-y", 1264),
-	PROPERTY_ENTRY_BOOL("touchscreen-inverted-y"),
-	PROPERTY_ENTRY_STRING("firmware-name", "/*(DEBLOBBED)*/"),
-	PROPERTY_ENTRY_U32("silead,max-fingers", 10),
-	PROPERTY_ENTRY_BOOL("silead,home-button"),
-	{ }
-};
-
-static const struct ts_dmi_data teclast_x16_plus_data = {
-	.embedded_fw = {
-		.name	= "/*(DEBLOBBED)*/",
-		.prefix = { 0xf0, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00 },
-		.length	= 43560,
-		.sha256	= { 0x9d, 0xb0, 0x3d, 0xf1, 0x00, 0x3c, 0xb5, 0x25,
-			    0x62, 0x8a, 0xa0, 0x93, 0x4b, 0xe0, 0x4e, 0x75,
-			    0xd1, 0x27, 0xb1, 0x65, 0x3c, 0xba, 0xa5, 0x0f,
-			    0xcd, 0xb4, 0xbe, 0x00, 0xbb, 0xf6, 0x43, 0x29 },
-	},
-	.acpi_name	= "MSSL1680:00",
-	.properties	= teclast_x16_plus_props,
-};
-
 static const struct property_entry teclast_x3_plus_props[] = {
 	PROPERTY_ENTRY_U32("touchscreen-size-x", 1980),
 	PROPERTY_ENTRY_U32("touchscreen-size-y", 1500),
@@ -1208,15 +1115,6 @@ const struct dmi_system_id touchscreen_dmi_table[] = {
 			DMI_MATCH(DMI_BIOS_VERSION, "CHUWI.D86JLBNR"),
 		},
 	},
-	{
-		/* Chuwi Vi8 dual-boot (CWI506) */
-		.driver_data = (void *)&chuwi_vi8_data,
-		.matches = {
-			DMI_MATCH(DMI_SYS_VENDOR, "Insyde"),
-			DMI_MATCH(DMI_PRODUCT_NAME, "i86"),
-			DMI_MATCH(DMI_BIOS_VERSION, "CHUWI2.D86JHBNR02"),
-		},
-	},
 	{
 		/* Chuwi Vi8 Plus (CWI519) */
 		.driver_data = (void *)&chuwi_vi8_plus_data,
@@ -1287,14 +1185,6 @@ const struct dmi_system_id touchscreen_dmi_table[] = {
 			DMI_MATCH(DMI_PRODUCT_NAME, "7W"),
 		},
 	},
-	{
-		/* DEXP Ursus KX210i */
-		.driver_data = (void *)&dexp_ursus_kx210i_data,
-		.matches = {
-			DMI_MATCH(DMI_SYS_VENDOR, "INSYDE Corp."),
-			DMI_MATCH(DMI_PRODUCT_NAME, "S107I"),
-		},
-	},
 	{
 		/* Digma Citi E200 */
 		.driver_data = (void *)&digma_citi_e200_data,
@@ -1376,17 +1266,6 @@ const struct dmi_system_id touchscreen_dmi_table[] = {
 			DMI_MATCH(DMI_BIOS_DATE, "04/24/2018"),
 		},
 	},
-	{
-		/* Jumper EZpad 6s Pro */
-		.driver_data = (void *)&jumper_ezpad_6_pro_b_data,
-		.matches = {
-			DMI_MATCH(DMI_SYS_VENDOR, "Jumper"),
-			DMI_MATCH(DMI_PRODUCT_NAME, "Ezpad"),
-			/* Above matches are too generic, add bios match */
-			DMI_MATCH(DMI_BIOS_VERSION, "/*(DEBLOBBED)*/"),
-			DMI_MATCH(DMI_BIOS_DATE, "01/08/2020"),
-		},
-	},
 	{
 		/* Jumper EZpad 6 m4 */
 		.driver_data = (void *)&jumper_ezpad_6_m4_data,
@@ -1416,18 +1295,6 @@ const struct dmi_system_id touchscreen_dmi_table[] = {
 			DMI_MATCH(DMI_BIOS_VERSION, "jumperx.T87.KFBNEEA"),
 		},
 	},
-	{
-		/* Juno Tablet */
-		.driver_data = (void *)&gdix1002_00_upside_down_data,
-		.matches = {
-			DMI_MATCH(DMI_SYS_VENDOR, "Default string"),
-			/* Both product- and board-name being "Default string" is somewhat rare */
-			DMI_MATCH(DMI_PRODUCT_NAME, "Default string"),
-			DMI_MATCH(DMI_BOARD_NAME, "Default string"),
-			/* Above matches are too generic, add partial bios-version match */
-			DMI_MATCH(DMI_BIOS_VERSION, "JP2V1."),
-		},
-	},
 	{
 		/* Mediacom WinPad 7.0 W700 (same hw as Wintron surftab 7") */
 		.driver_data = (void *)&trekstor_surftab_wintron70_data,
@@ -1549,14 +1416,6 @@ const struct dmi_system_id touchscreen_dmi_table[] = {
 			DMI_MATCH(DMI_BIOS_VERSION, "MOMO.G.WI71C.MABMRBA02"),
 		},
 	},
-	{
-		/* Positivo C4128B */
-		.driver_data = (void *)&positivo_c4128b_data,
-		.matches = {
-			DMI_MATCH(DMI_SYS_VENDOR, "Positivo Tecnologia SA"),
-			DMI_MATCH(DMI_PRODUCT_NAME, "C4128B-1"),
-		},
-	},
 	{
 		/* Point of View mobii wintab p800w (v2.0) */
 		.driver_data = (void *)&pov_mobii_wintab_p800w_v20_data,
@@ -1618,17 +1477,6 @@ const struct dmi_system_id touchscreen_dmi_table[] = {
 			DMI_MATCH(DMI_PRODUCT_SKU, "0001")
 		},
 	},
-	{
-		/* RWC NANOTE NEXT */
-		.driver_data = (void *)&rwc_nanote_next_data,
-		.matches = {
-			DMI_MATCH(DMI_PRODUCT_NAME, "To be filled by O.E.M."),
-			DMI_MATCH(DMI_BOARD_NAME, "To be filled by O.E.M."),
-			DMI_MATCH(DMI_BOARD_VENDOR, "To be filled by O.E.M."),
-			/* Above matches are too generic, add bios-version match */
-			DMI_MATCH(DMI_BIOS_VERSION, "S8A70R100-V005"),
-		},
-	},
 	{
 		/* Schneider SCT101CTM */
 		.driver_data = (void *)&schneider_sct101ctm_data,
@@ -1637,15 +1485,6 @@ const struct dmi_system_id touchscreen_dmi_table[] = {
 			DMI_MATCH(DMI_PRODUCT_NAME, "SCT101CTM"),
 		},
 	},
-	{
-		/* GlobalSpace SoLT IVW 11.6" */
-		.driver_data = (void *)&globalspace_solt_ivw116_data,
-		.matches = {
-			DMI_MATCH(DMI_SYS_VENDOR, "Globalspace Tech Pvt Ltd"),
-			DMI_MATCH(DMI_PRODUCT_NAME, "SolTIVW"),
-			DMI_MATCH(DMI_PRODUCT_SKU, "PN20170413488"),
-		},
-	},
 	{
 		/* Techbite Arc 11.6 */
 		.driver_data = (void *)&techbite_arc_11_6_data,
@@ -1664,15 +1503,6 @@ const struct dmi_system_id touchscreen_dmi_table[] = {
 			DMI_MATCH(DMI_PRODUCT_SKU, "E5A6_A1"),
 		},
 	},
-	{
-		/* Teclast X16 Plus */
-		.driver_data = (void *)&teclast_x16_plus_data,
-		.matches = {
-			DMI_MATCH(DMI_SYS_VENDOR, "TECLAST"),
-			DMI_MATCH(DMI_PRODUCT_NAME, "Default string"),
-			DMI_MATCH(DMI_PRODUCT_SKU, "D3A5_A1"),
-		},
-	},
 	{
 		/* Teclast X3 Plus */
 		.driver_data = (void *)&teclast_x3_plus_data,
@@ -1847,7 +1677,7 @@ static void ts_dmi_add_props(struct i2c_client *client)
 	int error;
 
 	if (has_acpi_companion(dev) &&
-	    strstarts(client->name, ts_data->acpi_name)) {
+	    !strncmp(ts_data->acpi_name, client->name, I2C_NAME_SIZE)) {
 		error = device_create_managed_software_node(dev, ts_data->properties, NULL);
 		if (error)
 			dev_err(dev, "failed to add properties: %d\n", error);
